# encoding: utf-8

module AppStack
  # items as copy list candidate
  class CompareFile
    attr_accessor :from_app, :from_file, :to_app, :to_file,
                  :import, :render, :diff
    def initialize(opts)
      opts.each { |k, v| send("#{k}=".to_sym, v) }
      @to_file ||= to_app.get_file_path(from_file_short_name)
    end

    def reverse!
      fail 'can not reverse an import or render item' if import? || render?
      from_app_, from_file_ = from_app.dup, from_file.dup
      @from_app, @from_file = to_app, to_file
      @to_app, @to_file = from_app_, from_file_
    end

    def render?
      render ? true : false
    end

    def import?
      import ? true : false
    end

    def sync?
      import ? false : true
    end

    def operation
      return 'render' if render?
      import? ? 'import' : 'sync'
    end

    def diff?
      return true unless File.exists?(to_file)
      c_file = from_file
      if render?
        c_file = Tempfile.new(File.basename(from_file)).path
        render! c_file
      end

      @diff ||= Diffy::Diff.new(to_file, c_file, source: 'files')
      @diff && @diff.to_s.chomp.size > 0 ? true : false
    end

    def local_new?
      File.exists?(to_file) && File.mtime(to_file) > File.mtime(from_file)
    end

    def remote_new?
      return true unless File.exists?(to_file)
      File.mtime(from_file) > File.mtime(to_file)
    end

    def from_file_short_name
      from_file.sub(/^#{from_app.directory}\//, '')
    end

    def to_file_short_name
      to_file.sub(/^#{to_app.directory}\//, '')
    end

    def label
      str = "#{operation} #{from_file_short_name.blue} from " +
            "#{from_app.app_name.bold}"
      str << " as #{to_file_short_name.blue}" unless
        to_file_short_name == from_file_short_name
      str
    end

    def reverse_label
      "copy back #{from_file_short_name.blue} to #{to_app.app_name.bold}"
    end

    def skip(msg)
      "Skip #{operation} ".white + from_file_short_name.blue + ' from '.white +
      from_app.app_name.bold + ': ' + msg
    end

    def process
      puts label
      render? ? render! : copy!
    end

    private

    def render!(to = nil)
      to ||= to_file
      # return if @options.simulate
      oh = File.open(to, 'wb')
      if from_file.match(/\.liquid$/)
        oh.write Liquid::Template
          .parse(File.open(from_file, 'r:utf-8').read).render(to_app.attrs)
      else
        tilt = Tilt.new(from_file)
        oh.write tilt.render(OpenStruct.new(to_app.attrs))
      end
      oh.close
    end

    def copy!
      target_dir = File.dirname(to_file)
      FileUtils.mkdir_p target_dir unless File.directory?(target_dir)
      FileUtils.cp_r from_file, to_file
    end
  end

  # reopen app class, extend with load_compare list function
  class App
    def load_compare_list!
      # @import_list.merge(@sync_list).merge(@render_list
      load_copy_list!(@import_list, import: true)
      load_copy_list!(@sync_list)
      load_render_list!(@render_list, render: true)
      load_local_render_list!(render: true)
    end

    def load_copy_list!(list, opts = {})
      list.each do |app, exp|
        exp.each do |p| # string as group name or hash as file mapping
          opts.merge! from_app: @app_stacks[app.to_s], to_app: self
          p.is_a?(String) ? load_group!(p, opts) : load_file_map!(p, opts)
        end
      end
    end

    # rubocop:disable MethodLength
    def load_render_list!(list, opts = {})
      list.each do |appname, hsh|
        app = @app_stacks[appname.to_s]
        hsh.each do |fr, to|
          to = fr.sub(/\.(erb|liquid|slim)$/, '') unless to.size > 0
          from_file = app.get_file_path(fr)
          if File.exists?(from_file)
            @compare_list << CompareFile.new(
              opts.merge(from_app: app, to_app: self,
                         from_file: from_file,
                         to_file: get_file_path(to)))
          else
            warn "[WARN] #{fr.blue} not found in #{app.app_name.bold}, skip."
          end
        end
      end
    end

    def load_local_render_list!(opts = {})
      config.render_local.each do |fr, to|
        from_file = get_file_path(fr)
        to = fr.sub(/\.(erb|liquid|slim)$/, '') unless to.size > 0
        if File.exists?(from_file)
          @compare_list << CompareFile.new(
            opts.merge(from_app: self, to_app: self,
                       from_file: from_file, to_file: get_file_path(to)))
        else
          warn "[WARN] #{fr.blue} not found in #{app_name.bold} (local), skip."
        end
      end
    end

    # rubocop:disable LineLength
    def load_group!(group, opts)
      app = opts[:from_app]
      fail ParseError, "Error on #{@conf_file}, #{app.app_name.bold} did not export group #{group.bold}" unless app.export_groups[group.to_s]
      app.get_group_files(group, opts[:from_app]).each do |full_path|
        @compare_list << CompareFile.new(opts.merge(from_file: full_path))
      end
    end

    def load_file_map!(hsh, opts)
      hsh.each do |f, t|
        t = f unless t && t.size > 0
        from_file = opts[:from_app].get_file_path(f)
        if File.exists?(from_file)
          @compare_list << CompareFile.new(
            opts.merge(from_file: opts[:from_app].get_file_path(f),
                       to_file: opts[:to_app].get_file_path(t)))
        else
          warn "[WARN] #{f.blue} not found in #{opts[:from_app].app_name.bold}, skip."
        end
      end
    end

    def get_group_files(group, app)
      file_list = []
      export_groups[group.to_s].each do |fp|
        if fp.match(/[\?\*\+\{\}]+/)
          Dir[get_file_path(fp)].each { |f| file_list << f }
        else
          file = get_file_path(fp)

          if File.exists?(file)
            file_list << file
          else
            warn "[WARN] #{fp.blue} not found for #{app.app_name.bold}, group #{group.bold}, skip."
          end
        end
      end
      file_list
    end
    # rubocop:enable LineLength, MethodLength

    def get_file_path(fp)
      File.expand_path(fp, directory)
    end
  end
end
